/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This is unpublished source code from BalaBit IT Ltd, and is trade secret. 
 * As such, it must be protected by all possible means to prevent
 * unauthorized access.
 *
 * $Id: authconn.c,v 1.5 2000/08/08 14:39:00 bazsi Exp $
 *
 ***************************************************************************/

#include <string.h>

#include <zorp/sockaddr.h>
#include <zorp/authconn.h>
#include <zorp/log.h>

#define Z_AUTHCONN_RETRY(a, rv) \
{ \
	z_warning(0, "z_authconn_retry(): error in authserver connection, retrying"); \
	if (!z_authconn_connect(a)) \
		return rv; \
	else \
		continue; } \


SSL_CTX *z_authconn_ctx = NULL;
SSL_METHOD *z_authconn_method = NULL;

int
z_authconn_write(ZAuthConnection * self, void *buf, int count)
{
  int err;
  switch (self->type)
    {
    case ZAC_TCP:
      return write(self->sock, buf, count);
    case ZAC_SSL:
      err = SSL_write(self->ssl, buf, count);
      return (err < 0) ? -1 : err;
    default:
      return -1;
    }
}

int
z_authconn_read(ZAuthConnection * self, void *buf, int count)
{
  switch (self->type)
    {
    case ZAC_TCP:
      return read(self->sock, buf, count);
    case ZAC_SSL:
      return SSL_read(self->ssl, buf, count);
    default:
      return -1;
    }
}

int
z_authconn_init(char *cert, char *priv)
{
  SSLeay_add_all_algorithms();
  z_authconn_method = SSLv3_client_method();
  SSL_load_error_strings();
  z_authconn_ctx = SSL_CTX_new(z_authconn_method);
  SSL_CTX_set_verify(z_authconn_ctx, 0, NULL);

  if (SSL_CTX_use_certificate_file(z_authconn_ctx, cert,
				   SSL_FILETYPE_PEM) <= 0)
    {
      SSL_CTX_free(z_authconn_ctx);
      return 0;
    }
  if (SSL_CTX_use_PrivateKey_file(z_authconn_ctx, priv,
				  SSL_FILETYPE_PEM) <= 0)
    {
      SSL_CTX_free(z_authconn_ctx);
      return 0;
    }

  if (!SSL_CTX_check_private_key(z_authconn_ctx))
    {
      SSL_CTX_free(z_authconn_ctx);
      return 0;
    }

  return 1;
}

int
z_authconn_readline(ZAuthConnection * self, char *buf, int len)
{
  char ch;
  int err;
  int ctr = 0;

  while (ctr < len)
    {
      err = z_authconn_read(self, &ch, 1);
      if (err != 1)
	return -1;
      if (ch == '\n')
	break;
      buf[ctr++] = ch;
    }
  buf[ctr] = 0;
  return strlen(buf);
}

int
z_authconn_connect_tcp(ZAuthConnection * self)
{
  char buf[1024];

  if (self->type != ZAC_TCP ||
      self->ssl != NULL || self->peername != NULL || self->myname == NULL)
    return 0;

  if (self->sock != -1)
    {
      if (self->sock > 0)
	close(self->sock);
      self->sock = -1;
    }

  if (strlen(self->myname) > 1022)
    goto bad;

  self->sock = socket(AF_INET, SOCK_STREAM, 0);
  if (self->sock < 0)
    goto bad;
  if (z_connect(self->sock, self->sa) != G_IO_ERROR_NONE)
    goto bad;

  snprintf(buf, 1023, "%s\n", self->myname);
  if (write(self->sock, buf, strlen(buf)) != strlen(buf))
    goto bad;

  buf[5] = 0;
  if (read(self->sock, buf, 5) != 5 || strcmp(buf, "YEAH\n"))
    goto bad;

  return 1;

bad:
  if (self->sock != -1)
    close(self->sock);
  return 0;
}

int
z_authconn_connect_ssl(ZAuthConnection * self)
{
  int err, connected = 0;
  char *server_dn = NULL;
  X509 *server_cert = NULL;
  char buf[6];

  if (self->type != ZAC_SSL || self->myname != NULL)
    return 0;

  if (z_authconn_ctx == NULL)
    goto bad;

  if (self->ssl != NULL)
    {
      SSL_shutdown(self->ssl);
      SSL_free(self->ssl);
      self->ssl = NULL;
    }

  if (self->sock != -1)
    {
      if (self->sock > 0)
	close(self->sock);
      self->sock = -1;
    }

  self->sock = socket(AF_INET, SOCK_STREAM, 0);
  if (self->sock < 0)
    goto bad;
  if (z_connect(self->sock, self->sa) != G_IO_ERROR_NONE)
    goto bad;
  
  self->ssl = SSL_new(z_authconn_ctx);
  SSL_set_fd(self->ssl, self->sock);
  err = SSL_connect(self->ssl);
  if (err < 0)
    goto bad;
  connected = -1;

  server_cert = SSL_get_peer_certificate(self->ssl);
  server_dn = X509_NAME_oneline(X509_get_subject_name(server_cert), 0, 0);
  if (server_dn == NULL)
    {
      z_message(0, "z_authconn_connect_ssl(): no server dn, disconnecting");
      goto bad;
    }

  z_message(0, "z_authconn_connect_ssl(): server dn=%s", server_dn);
  if (self->peername != NULL && strcasecmp(self->peername, server_dn))
    goto bad;

  buf[5] = 0;
  if (SSL_read(self->ssl, buf, 5) != 5 || strcmp(buf, "YEAH\n"))
    goto bad;

  CRYPTO_free(server_dn);
  X509_free(server_cert);
  return 1;
bad:
  if (server_dn != NULL)
    free(server_dn);
  if (server_cert != NULL)
    X509_free(server_cert);
  if (connected)
    SSL_shutdown(self->ssl);
  if (self->ssl != NULL)
    SSL_free(self->ssl);
  if (self->sock != -1)
    close(self->sock);
  return 0;
}


int
z_authconn_connect(ZAuthConnection * self)
{
  switch (self->type)
    {
    case ZAC_TCP:
      return z_authconn_connect_tcp(self);
    case ZAC_SSL:
      return z_authconn_connect_ssl(self);
    default:
      return 0;
    }
}

ZAuthConnection *
z_authconn_new_tcp(ZSockAddr * sa, char *myname)
{
  ZAuthConnection *self = g_new0(ZAuthConnection, 1);

  if (strlen(myname) > 1022)
    {
      g_free(self);
      return NULL;
    }

  self->sock = -1;
  self->myname = strdup(myname);
  self->sa = z_sockaddr_ref(sa);
  self->peername = NULL;
  self->ssl = NULL;
  self->type = ZAC_TCP;

  if (!z_authconn_connect_tcp(self))
    {
      z_authconn_destroy(self);
      return NULL;
    }

  return self;
}

ZAuthConnection *
z_authconn_new_ssl(ZSockAddr * sa, char *peername)
{
  ZAuthConnection *self = g_new0(ZAuthConnection, 1);

  self->type = ZAC_SSL;
  self->sock = -1;
  self->ssl = NULL;
  self->myname = NULL;
  self->sa = z_sockaddr_ref(sa);

  if (peername != NULL)
    self->peername = strdup(peername);
  else
    self->peername = NULL;

  if (!z_authconn_connect_ssl(self))
    {
      z_authconn_destroy(self);
      return NULL;
    }
  return self;

}

void
z_authconn_destroy_ssl(ZAuthConnection * self)
{
  if (self->ssl != NULL) 
    {
      SSL_shutdown(self->ssl);
      SSL_free(self->ssl);
    }
  if (self->sock != -1)
    {
      close(self->sock);
    }
  g_free(self);
}

void
z_authconn_destroy_tcp(ZAuthConnection * self)
{
  close(self->sock);
  g_free(self);
}

void
z_authconn_destroy(ZAuthConnection * self)
{
  if (self->peername != NULL)
    free(self->peername);
  if (self->myname != NULL)
    free(self->myname);
  z_sockaddr_unref(self->sa);

  switch (self->type)
    {
    case ZAC_TCP:
      z_authconn_destroy_tcp(self);
      break;
    case ZAC_SSL:
      z_authconn_destroy_ssl(self);
      break;
    }
}

char **
z_authconn_parse_types(ZAuthConnection * self, int *count)
{
  int i, idx;
  int err;
  char **result;
  char buf[1024];

  err = z_authconn_readline(self, buf, 1023);
  if (err == -1)
    return NULL;

  *count = (strlen(buf) + 1) / 5;
  result = g_new(char *, *count + 1);
  result[*count] = NULL;

  for (i = idx = 0; i < *count; i++)
    {
      result[i] = g_new(char, 5);
      strncpy(result[i], &buf[i * 5], 4);
      result[i][4] = 0;
    }

  return result;
}

void
z_authconn_free(char **what)
{
  int i;

  for (i = 0; what[i] != NULL; i++)
    g_free(what[i]);

  g_free(what);
}

char **
z_authconn_type_int(ZAuthConnection * self, int *count, int retry)
{
  int i;
  int err;
  char buf[1024];

  for (i = 0; i <= retry; i++)
    {
      err = z_authconn_write(self, "TYPE\n", 5);
      if (err == -1)
	Z_AUTHCONN_RETRY(self, NULL)

      err = z_authconn_readline(self, buf, 1023);
      if (err == -1)
	Z_AUTHCONN_RETRY(self, NULL)

      if (strcmp(buf, "YEAH"))
	return NULL;

      return z_authconn_parse_types(self, count);
    }
  return NULL;
}

char *
z_authconn_escape(char *o, char *i, int len)
{
  int ctr = 0;

  while (*i && ctr < len)
    {
      if (*i == '"')
	{
	  o[ctr++] = '"';
	  o[ctr++] = '"';
	}
      else
	o[ctr++] = *i;
      i++;
    }
  o[ctr] = 0;
  return o;
}

char **
z_authconn_utyp_int(ZAuthConnection * self, char *dn, int *count, int retry)
{
  int i;
  int err;
  char buf[1024];
  char escape_buf[1024];

  for (i = 0; i <= retry; i++)
    {
      z_authconn_escape(escape_buf, dn, 1024);
      snprintf(buf, 1023, "UTYP \"%s\"\n", escape_buf);
      err = z_authconn_write(self, buf, strlen(buf));
      if (err == -1)
	Z_AUTHCONN_RETRY(self, NULL)
		
      err = z_authconn_readline(self, buf, 1023);
      if (err == -1)
	Z_AUTHCONN_RETRY(self, NULL)

      if (strcmp(buf, "YEAH"))
	return NULL;
      return z_authconn_parse_types(self, count);
    }
  return NULL;
}

int
z_authconn_chal_int(ZAuthConnection * self, char *type, char *dn, char *out,
		    int len, int retry)
{
  int i;
  int err;
  char buf[1024];
  char escape_buf[1024];

  z_authconn_escape(escape_buf, dn, 1023);

  for (i = 0; i <= retry; i++)
    {
      snprintf(buf, 1023, "CHAL %s \"%s\"\n", type, escape_buf);
      err = z_authconn_write(self, buf, strlen(buf));
      if (err == -1)
	Z_AUTHCONN_RETRY(self, 0)

      err = z_authconn_readline(self, buf, 1023);
      if (err == -1)
	Z_AUTHCONN_RETRY(self, 0)

      if (strcmp(buf, "YEAH"))
	return 0;

      err = z_authconn_readline(self, buf, 1023);
      if (err == -1 || err >= len - 2)
	Z_AUTHCONN_RETRY(self, 0)

      strcpy(out, buf);
      return 1;
    }

  return 0;

}

int
z_authconn_parse_groups(char *from, char *to[], int *ctr)
{
  char buf[1024], *bp;
  char *p = from;

  *ctr = 0;
  while (*p != '"' && *p)
    p++;

  while (*p)
    {
      p++;
      bp = buf;
      while (*p && (*p != '"' || (*p == '"' && p[1] == '"')))
	{
	  if (*p != '"')
	    *bp++ = *p++;
	  else
	    *bp++ = '"', p++, p++;
	}
      if (!*p)
	goto bad_end;
      *bp = 0;
      to[(*ctr)++] = strdup(buf);
      p++;
      while (*p != '"' && *p)
	p++;
    }
  return 1;
bad_end:
  while (*ctr)
    {
      (*ctr)--;
      free(to[*ctr]);
    }
  return 1;
}

ZLabelSet *
z_authconn_auth_int(ZAuthConnection * self, char *typename,
		    char *dn, char *chal, char *resp, int retry)
{
  int i;
  int err;
  char buf[1024];
  char chal_buf[1024], dn_buf[1024], resp_buf[1024];
  ZLabelSet *label_set;

  if (typename == NULL)
    return NULL;

  z_authconn_escape(chal_buf, chal, 1023);
  z_authconn_escape(dn_buf, dn, 1023);
  z_authconn_escape(resp_buf, resp, 1023);

  for (i = 0; i <= retry; i++)
    {
      snprintf(buf, 1023, "AUTH %s \"%s\" \"%s\" \"%s\"\n",
	       typename, chal_buf, dn_buf, resp_buf);
      err = z_authconn_write(self, buf, strlen(buf));
      if (err == -1)
	Z_AUTHCONN_RETRY(self, NULL);

      err = z_authconn_readline(self, buf, 1023);
      if (err == -1)
	Z_AUTHCONN_RETRY(self, NULL);
      if (strcmp(buf, "YEAH"))
	return NULL;

      if (z_authconn_readline(self, buf, 1023) == -1)
	Z_AUTHCONN_RETRY(self, NULL);

      label_set = z_label_set_new();
      if (sscanf(buf, "%d %d", &label_set->rmax, &label_set->wmin) != 2)
	{
	  z_label_set_unref(label_set);
	  return NULL;
	}
      if (z_authconn_readline(self, buf, 1023) == -1)
	{
	  z_label_set_unref(label_set);
	  Z_AUTHCONN_RETRY(self, NULL);
	}
      if (!z_authconn_parse_groups(buf, label_set->nhcs, &label_set->nhc_ctr))
	{
	  z_label_set_unref(label_set);
	  return NULL;
	}
      if (z_authconn_readline(self, buf, 1023) == -1)
	{
	  z_label_set_unref(label_set);
	  Z_AUTHCONN_RETRY(self, NULL);
	}
      if (!z_authconn_parse_groups(buf,
				   label_set->groups, &label_set->group_ctr))
	{
	  z_label_set_unref(label_set);
	  return NULL;
	}
      return label_set;
    }
  return NULL;
}

ZLabelSet *
z_authconn_auth(ZAuthConnection * ac,
		char *type, char *dn, char *chal, char *resp)
{
  return z_authconn_auth_int(ac, type, dn, chal, resp, 1);
}

int
z_authconn_chal(ZAuthConnection * ac,
		char *type, char *dn, char *buf, int len)
{
  return z_authconn_chal_int(ac, type, dn, buf, len, 1);
}

char **
z_authconn_type(ZAuthConnection * ac, int *count)
{
  return z_authconn_type_int(ac, count, 1);
}

char **
z_authconn_utyp(ZAuthConnection * ac, char *dn, int *count)
{
  return z_authconn_utyp_int(ac, dn, count, 1);
}
