/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ftpparse.c,v 1.8 2000/08/08 16:48:57 bazsi Exp $
 *
 * Author:  Andras Kis-Szabo <kisza@sch.bme.hu>
 * Auditor:
 * Notes:
 *
 ***************************************************************************/

#include <zorp/zorp.h>
#include <zorp/stream.h>
#include <zorp/readline.h>
#include <zorp/proxy.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/zpython.h>
#include <zorp/log.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>


#include "ftp.h"
#include "ftphash.h"
#include "ftpmain.h"
#include "ftpparse.h"

char *msgs[] = {
  "501 Cannot parse PORT command.",
  "502 REIN command not implemented.",
  "500 Command disabled.",
  "501 Password/account format invalid.",
  "503 Please login with USER and PASS.",
  "503 Login with USER first.",
  "530 Already logged in.",
  "530 User access not allowed.",
  "501 Username format invalid.",
  "502 Bad ALLO argument.",

  "501 Pathname format invalid.",
  "501 Marker format invalid.",
  "501 Argumentum format invalid.",
  "502 Command disabled.",
  "501 Too many parameters.",
  "501 Cannot parse TYPE command.",
  "504 Byte size must be 8.",
  "504 Form must be N.",
  "504 Type E not implemented.",
  "501 Invalid STRU parameter.",

  "502 Unimplemented STRU type.",
  "501 Invalid MODE parameter.",
  "502 Unimplemented MODE type.",
  "220- Welcome in Zorp FTP proxy authentication module!\r\n220- Please authenticate yourself!\r\n220- Authetication form:\r\n220-  USER <user>@<remote site>\r\n220-  PASS <password>\r\n220 ",
  "500 Line termination must be CR LF.",
  "500 Command line too long.",
  "500 Syntax error, command unrecognized.",
  "421 Service not available, remote server has closed connection.",
  "425 Can't build data connection: bounce attack detected." "200 200 PORT command successful."
};


//#define ENABLE_DEBUG

// Function:    ftp_answerstate
// In:          gpointer                proxy pointer
// Out:         -
// Desc:        helper function, used in server answer state-machine
//              for state-changing. (Once in all answer-group)
void
ftp_answerstate (gpointer s)
{
  FtpProxy *self = (FtpProxy *) s;

  switch (self->ftp_state)
    {
    case FTP_STATE_CONNECT:
      self->ftp_state = FTP_STATE_LOGIN;
      break;
    default:
      break;
    }
}

// Function:    ftp_strescape
// In:          gchar *                 line
// Out:         gchar *                 escaped IAC
// Desc:        Escaping the IAN in NVT protocoll
gchar *
ftp_strescape (const gchar * source)
{
  const guchar *work_source = (guchar *) source;
  /* Each source byte needs maximally two destination chars */
  gchar *dest = g_malloc (strlen (source) * 2 + 1);
  gchar *work_dest = dest;

  while (*work_source)
    {
      switch (*work_source)
	{
	case 255:
	  *work_dest++ = 255;
	  *work_dest++ = 255;
	  break;
	default:
	  *work_dest++ = *work_source;
	  break;
	}
      work_source++;
    }
  *work_dest = 0;

  return dest;
}

// Function:    ftp_stream_write
// In:          FtpProxy *              proxy pointer
//              char                    direction S C
//              gchar *                 command
//              int                     length
// Out:         int                     sended bytes
// Desc:        Send the command to the server
guint
ftp_stream_write (FtpProxy * self, char dir, char *line, int length)
{
  guint bytes_written;
  gchar *buff;
  gchar *msg;
  int back;
#ifdef ENABLE_DEBUG
  gchar *buff2;
#endif

  buff = g_new0 (char, length + 3);

#ifdef ENABLE_DEBUG
  buff2 = g_new0 (char, length + 3);
#endif

  snprintf (buff, length + 3, "%.*s\r\n", length, line);

#ifdef ENABLE_DEBUG
  snprintf (buff2, length + 3, "%.*srn", length, line);
  z_debug (7, "%s: FTP: write --%s-- ", self->super.session_id, buff2);
#endif

  msg = ftp_strescape (buff);

  switch (dir)
    {
    case 'S':
      z_stream_write (self->super.endpoints[EP_SERVER], msg, strlen (msg), &bytes_written);
      break;
    case 'C':
      z_stream_write (self->super.endpoints[EP_CLIENT], msg, strlen (msg), &bytes_written);
      break;
    default:
      break;
    }

#ifdef ENABLE_DEBUG
  if (buff2)
    g_free (buff2);
#endif

  back = strlen (msg);

  if (buff)
    g_free (buff);

  if (msg)
    g_free (msg);

  //return MIN (bytes_written, back);
  if ( bytes_written == back ) 
    {
      return 0;
    }
  else
    {
      return 1; 
    }
}

// Function:    ftp_log
// In:          gpointer                proxy pointer
//              gint                    log-flags
//              gchar *                 message for send
//              int                     length 
// Out:         int                     success?
// Desc:        log sender function.
//              Send an event with the flags and the message
int
ftp_log (gpointer s, gint from, gchar * msg, int length)
{
  FtpProxy *self = (FtpProxy *) s;
  char *what[] = { "command", "answer" };
  char *done[] = { "dropped", "passed" };
  char *src[] = { "protocol", "proxy" };

  z_message (7, "%s: %s %s from %s: %.*s", self->super.session_id, done[((unsigned int) from & LOG_PASS) >> 4], what[((unsigned int) from & LOG_ANSW) >> 1], src[((unsigned int) from & LOG_PROX) >> 5], length, msg);

  return 1;
}

// Function:    ftp_sendback
// In:          FtpProxy *              proxy pointer
//              gchar *                 message for send
// Out:         int                     sended bytes
// Desc:        Answering sender function.
//              Send the message back to the client
int
ftp_sendback (FtpProxy * self, gchar * msg)
{
  guint bytes_written;

  if (self->verbose_messages)
    {
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: sendback --%s-- ", self->super.session_id, msg);
#endif
      bytes_written = ftp_stream_write (self, 'C', msg, strlen (msg));
    }
  else
    {
      bytes_written = ftp_stream_write (self, 'C', msg, 4);
    }

  if (bytes_written)
    {
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: failed to send \"%s\"", self->super.session_id, msg);
#endif
      self->quit = 1;
      self->state = FTP_MAIN_RET;
    }

  return bytes_written;
  // checked after return
}

// Function:    ftp_command_getCC
// In:          long                    command flag
// Out:         char                    CC flag
// Desc:        Generate CC type for event
char
ftp_command_getCC (unsigned long flag)
{
  char help='U';

  switch (flag & 0x0000000FL)
    {
    case 8:
      help = 'R';
      break;
    case 4:
      help = 'W';
      break;
    case 2:
      help = 'C';
      break;
    case 1:
      help = 'F';
      break;
    default:
      help = 'U';
      break;
    }

  return help;
}

// Function:    ftp_command_flag
// In:          FtpProxy                self
//              long                    command flag
//              long                    command mask
// Out:         boolean                 evaluated
// Desc:        Makes flag ...
gboolean
ftp_command_flag (FtpProxy * self, int mode, unsigned long flag, unsigned long mask)
{
  int help;

  // rfc    -> 0xFF000000L
  // level  -> 0x00FFF000L
  // access -> 0x00000F80L
  // name   -> 0x00000070L
  // CC     -> 0x0000000FL

  switch (mode)
    {
    case 0:
      // Command
      help = (((flag & mask) & 0xFF000000L) && ((flag & mask) & 0x00FFF000L) && ((flag & mask) & 0x00000F80L) && ((flag & mask) & 0x00000070L) && ((flag & mask) & 0x0000000FL)) || ftp_local_hash_get (self, self->command);
      break;
    case 1:
      // Logging
      help = ((flag & mask) & 0xFF000000L) && ((flag & mask) & 0x00FFF000L) && ((flag & mask) & 0x00000F80L) && ((flag & mask) & 0x00000070L) && ((flag & mask) & 0x0000000FL);
      break;
    default:
      break;
    }

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: FTP: mask: %08lX %08lX %08lX %08lX %08lX %d back: %d", self->super.session_id,
	   ((flag & mask) & 0xFF000000L), ((flag & mask) & 0x00FFF000L), ((flag & mask) & 0x00000F80L), ((flag & mask) & 0x00000070L), ((flag & mask) & 0x0000000FL), ftp_local_hash_get (self, self->command), help);
#endif

  return help;
}


// Function:    ftp_command_log
// In:          FtpProxy *              proxy pointer
//              gchar *                 command
//              int                     length
//              unsigned long           flags
// Out:         int                     sended bytes
// Desc:        Send the command to the server, and
//              logging it, if needed
guint
ftp_command_log (FtpProxy * self, char *line, int length, unsigned long active_flag)
{
  guint bytes_written;
  bytes_written = ftp_stream_write (self, 'S', line, length);
  if (ftp_command_flag (self, 1, active_flag, self->command_log))
    {
      while (!ftp_log (self, LOG_COMM + LOG_PASS, line, length));
    }
  return bytes_written;
  // checked after return
}

// Function:    ftp_command_drop
// In:          FtpProxy *              proxy pointer
//              gchar *                 command
//              int                     length
//              gchar *                 message
// Out:         int                     sended bytes
// Desc:        Dropping a command:
//              - send back the rejecting message
//              - log the dropped command
//              - log the sended message
guint
ftp_command_drop (FtpProxy * self, char *line, int length, char *msg)
{
  guint back;

  back = ftp_sendback (self, msg);
  if (self->log_dropped)
    {
      while (!ftp_log (self, LOG_COMM + LOG_DROP, line, length));
    }
  if (self->log_answer)
    {
      while (!ftp_log (self, LOG_ANSW + LOG_PASS + LOG_PROX, msg, strlen (msg)));
    }
  return back;
  // checked after return
}

// Function:    ftp_get_port
// In:          FtpProxy *              proxy pointer
//              char *                  ip number
//              int			side
// Out:         unsigned int            port info
// Desc:        Build standard connection after a PORT
unsigned int
ftp_get_port (FtpProxy * self,const char * ip,int side)
{
  unsigned int fakeport;
  ZPolicyObj *res;

  z_policy_lock (self->super.thread);
  res = z_policy_call (self->super.handler, "prepareDATA", z_policy_var_build ("si", ip, side));

  if (PyInt_Check(res))
    fakeport = PyInt_AsLong(res);
  else
    fakeport = 0;
  z_policy_var_unref(res);

  z_policy_unlock (self->super.thread);

  return fakeport;
}

void
ftp_test (FtpProxy * self)
{
  z_debug (9, "%s: fw_external   %s:%d", self->super.session_id, self->fw_external_data_address->str, self->fw_external_data_port);
  z_debug (9, "%s: fw_internal   %s:%d", self->super.session_id, self->fw_internal_data_address->str, self->fw_internal_data_port);
  z_debug (9, "%s: client        %s:%d", self->super.session_id, self->client_data_address->str, self->client_data_port);
  z_debug (9, "%s: server        %s:%d", self->super.session_id, self->server_data_address->str, self->server_data_port);
  z_debug (9, "%s: tmp           %s:%d", self->super.session_id, self->tmp_data_address->str, self->tmp_data_port);
}

// Function:    ftp_data_linear_PORT
// In:          FtpProxy *              proxy pointer
//              FTPCommandDescriptor *  command
// Out:         int                     
// Desc:        Build standard connection after a PORT
guint
ftp_data_linear_PORT (FtpProxy * self, FTPCommandDescriptor * command)
{

  int j;

  ftp_test (self);

  if (self->NAT)
    {
      if (self->internal_server)
        {
	   z_policy_lock (self->super.thread);
           j = z_policy_event(self->super.handler, "startDATA",
			  //
			  //  Connection sheme: 
			  //    Listen  client:client   server:*
			  //    Connect fw_e:*          client:client
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'L', 
self->client_data_address->str, self->client_data_port,
self->server_data_address->str, 0, 
'C', 
self->fw_external_data_address->str, 0, 
self->client_data_address->str, self->client_data_port));
	   z_policy_unlock (self->super.thread);
	}
      else
        {
	   z_policy_lock (self->super.thread);
           j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  fw_e:fw_e   server:*
			  //    Connect server:*    client:client
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'A', 
self->fw_external_data_address->str, self->fw_external_data_port,
self->server_data_address->str, 0, 
'C', 
self->server_data_address->str, 0, 
self->client_data_address->str, self->client_data_port));
	   z_policy_unlock (self->super.thread);
        }
    }
  else
    {
	   z_policy_lock (self->super.thread);
      j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  client:client server:*
			  //    Connect server:*      client:client
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'L', 
self->client_data_address->str, self->client_data_port,
self->server_data_address->str, 0,
'C', 
self->server_data_address->str, 0, 
self->client_data_address->str, self->client_data_port));
	   z_policy_unlock (self->super.thread);
    }

  return j;
}

// Function:    ftp_data_linear_PASV
// In:          FtpProxy *              proxy pointer
//              FTPCommandDescriptor *  command
// Out:         int                     
// Desc:        Build standard connection after a PASV
guint
ftp_data_linear_PASV (FtpProxy * self, FTPCommandDescriptor * command)
{

  int j;
  //ZPolicyObj *res;

  ftp_test (self);

  if (self->NAT)
    {
      if(self->internal_server)
        {
	   z_policy_lock (self->super.thread);
          j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  fw_e:fw_e client:*
			  //    Connect client:*  server:server
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'A', 
self->fw_external_data_address->str, self->fw_external_data_port, 
self->client_data_address->str, 0, 
'C', 
self->client_data_address->str, 0, 
self->server_data_address->str, self->server_data_port));
	   z_policy_unlock (self->super.thread);
	}
      else
        {
	   z_policy_lock (self->super.thread);
          j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme: direct connection
			  //    Listen  server:server client:*
			  //    Connect fw_e:*        server:server
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'L', 
self->server_data_address->str, self->server_data_port, 
self->client_data_address->str, 0, 
'C', 
self->fw_external_data_address->str, 0, 
self->server_data_address->str, self->server_data_port));
	   z_policy_unlock (self->super.thread);
        }
    }
  else
    {
	   z_policy_lock (self->super.thread);
      j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  server:server client:*
			  //    Connect client:*      server:server
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'L', 
self->server_data_address->str, self->server_data_port, 
self->client_data_address->str, 0, 
'C', 
self->client_data_address->str, 0, 
self->server_data_address->str, self->server_data_port));
	   z_policy_unlock (self->super.thread);
    }

  return j;
}

// Function:    ftp_data_linear_PORTPASV
// In:          FtpProxy *              proxy pointer
//              FTPCommandDescriptor *  command
// Out:         int                     
// Desc:        Build standard connection after a PORTPASV
guint
ftp_data_linear_PORTPASV (FtpProxy * self, FTPCommandDescriptor * command)
{

  int j;
  //ZPolicyObj *res;

  ftp_test (self);

  if (self->NAT)
    {
      if(self->internal_server)
        {
	   z_policy_lock (self->super.thread);
          j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Connect fw_e:*   client:client
			  //    Connect client:* server:server
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'C', 
self->fw_external_data_address->str, 0, 
self->client_data_address->str, self->client_data_port, 
'C', 
self->client_data_address->str, 0, 
self->server_data_address->str, self->server_data_port));
	   z_policy_unlock (self->super.thread);
	}
      else
        {
	   z_policy_lock (self->super.thread);
          j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Connect server:* client:client
			  //    Connect fw_e:*   server:server
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'C', 
self->server_data_address->str, 0, 
self->client_data_address->str, self->client_data_port, 
'C', 
self->fw_external_data_address->str, 0, 
self->server_data_address->str, self->server_data_port));
	   z_policy_unlock (self->super.thread);
        }
    }
  else
    {
	   z_policy_lock (self->super.thread);
      j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Connect server:* client:client
			  //    Connect client:* server:server
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'C', 
self->server_data_address->str, 0, 
self->client_data_address->str, self->client_data_port, 
'C', 
self->client_data_address->str, 0, 
self->server_data_address->str, self->server_data_port));
	   z_policy_unlock (self->super.thread);
    }

  return j;
}

// Function:    ftp_data_linear_PASVPORT
// In:          FtpProxy *              proxy pointer
//              FTPCommandDescriptor *  command
// Out:         int                     
// Desc:        Build standard connection after a PASVPORT
guint
ftp_data_linear_PASVPORT (FtpProxy * self, FTPCommandDescriptor * command)
{

  int j;
  //ZPolicyObj *res;

  ftp_test (self);

  if (self->NAT)
    {
      if(self->internal_server)
        {
	   z_policy_lock (self->super.thread);
          j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  client:client server:*
			  //    Listen  fw_e:fw_e     client:*
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'B', 
self->client_data_address->str, self->client_data_port, 
self->server_data_address->str, 0,
'A', 
self->fw_external_data_address->str, self->fw_external_data_port, 
self->client_data_address->str, 0));
	   z_policy_unlock (self->super.thread);
	}
      else
        {
	   z_policy_lock (self->super.thread);
          j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  fw_e:fw_e     server:*
			  //    Listen  server:server client:*
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'A', 
self->fw_external_data_address->str, self->fw_external_data_port, 
self->server_data_address->str, 0,
'B', 
self->server_data_address->str, self->server_data_port, 
self->client_data_address->str, 0));
	   z_policy_unlock (self->super.thread);
        }
    }
  else
    {
	   z_policy_lock (self->super.thread);
      j = z_policy_event (self->super.handler, "startDATA",
			  //
			  //  Connection sheme:
			  //    Listen  client:client server:*
			  //    Listen  server:server client:*
			  //
			  z_policy_var_build ("ccsisicsisi", ftp_command_getCC (command->flag), 'A', 
self->client_data_address->str, self->client_data_port, 
self->server_data_address->str, 0,
'B', 
self->server_data_address->str, self->server_data_port, 
self->client_data_address->str, 0));
	   z_policy_unlock (self->super.thread);
    }

  return j;
}

// Function:    ftp_data_channel
// In:          FtpProxy *              proxy pointer
//              FTPCommandDescriptor    command
// Out:         int                     state
// Desc:        Starts the data channel
guint
ftp_data_channel (FtpProxy * self, FTPCommandDescriptor * command)
{
	int j;

	  switch (self->client_portpasv)
	    {
	    case 0:
	      // PORT & PASV
	      if (self->last_port)
		{
		  j = ftp_data_linear_PORT (self, command);
		}
	      else
		{
		  j = ftp_data_linear_PASV (self, command);
		}
	      break;
	    case 1:
	      // PORT, PASV -> PORT
	      if (self->last_port)
		{
		  // PORT
		  j = ftp_data_linear_PORT (self, command);
		}
	      else
		{
		  // PASV -> PORT
		  j = ftp_data_linear_PASVPORT (self, command);
		}
	      break;
	    case 2:
	      // PORT -> PASV, PASV
	      if (self->last_port)
		{
		  // PORT -> PASV
		  j = ftp_data_linear_PORTPASV (self, command);
		}
	      else
		{
		  // PASV
		  j = ftp_data_linear_PASV (self, command);
		}
	      break;
	    default:
	          j=Z_ABORT;
	      break;
	    }
	  
	  return j;
}

// Function:    ftp_event
// In:          FtpProxy *              proxy pointer
//              int                     event mode
//              char *                  line
//              int                     length
//              FTPCommandDescriptor    command
// Out:         int                     state
// Desc:        Send an event
guint
ftp_event (FtpProxy * self, int mode, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  //int j;

  ftp_test (self);

  if (!ftp_command_flag (self, 0, command->flag, self->command_accept))
    {
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: %s not autoaccepted", self->super.session_id, command->name);
#endif
      z_policy_lock (self->super.thread);
      switch (mode)
	{
	case 0:
	case 1:
	case 4:
	  i = z_policy_event (self->super.handler, command->event, z_policy_var_build ("clss", ftp_command_getCC (command->flag), command->flag, self->command, self->parameter));
	  break;
	case 2:
	  i = z_policy_event (self->super.handler, command->event, z_policy_var_build ("cs", ftp_command_getCC (command->flag), self->parameter));
	  break;
	case 3:
	  i = z_policy_event (self->super.handler, command->event, z_policy_var_build ("(c)", ftp_command_getCC (command->flag)));
	  break;

	}
      z_policy_unlock (self->super.thread);
    }
  else
    {
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: %s autoaccepted", self->super.session_id, command->name);
#endif
      i = Z_ACCEPT;
    }
  switch (i)
    {
    case Z_ACCEPT:
      switch (mode)
	{
	case 0:
	case 2:
	case 3:
	case 4:
	  ftp_command_log (self, line, length, command->flag);
	  self->state = FTP_SERVER_TO_CLIENT;
	  break;
	case 1:
	  break;
	}

#if 0
      if (command->need_data)
	{
	  switch (self->client_portpasv)
	    {
	    case 0:
	      // PORT & PASV
	      if (self->last_port)
		{
		  j = ftp_data_linear_PORT (self, command);
		}
	      else
		{
		  j = ftp_data_linear_PASV (self, command);
		}
	      break;
	    case 1:
	      // PORT, PASV -> PORT
	      if (self->last_port)
		{
		  // PORT
		  j = ftp_data_linear_PORT (self, command);
		}
	      else
		{
		  // PASV -> PORT
		  j = ftp_data_linear_PASVPORT (self, command);
		}
	      break;
	    case 2:
	      // PORT -> PASV, PASV
	      if (self->last_port)
		{
		  // PORT -> PASV
		  j = ftp_data_linear_PORTPASV (self, command);
		}
	      else
		{
		  // PASV
		  j = ftp_data_linear_PASV (self, command);
		}
	      break;
	    default:
	      break;
	    }
	}
#endif
      break;
    case Z_DENY:
      break;
    case Z_UNSPEC:
    case Z_REJECT:
      switch (mode)
	{
	case 0:
	  ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	  break;
	case 1:
	case 2:
	case 3:
	case 4:
	  break;
	}
      break;
    case Z_ABORT:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }

  return i;
}

// Function:    ftp_command_parse_noarg
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        command with empty arg
int
ftp_command_parse_noarg (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((length - strlen (command->name) > 0))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_TOO_MANY_PARAM]);
	}
      else
	{
	  strcpy (self->parameter, "");
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter --", self->super.session_id, command->name);
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 0, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_PASV
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        command with empty arg
//  TODO: variants ( A, E, L )...
int
ftp_command_parse_PASV (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  unsigned int fakeport;
  char buf[29];

  ftp_test (self);

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((length - strlen (command->name) > 0))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_TOO_MANY_PARAM]);
	}
      else
	{
	  strcpy (self->parameter, "");
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter --", self->super.session_id, command->name);
#endif
	  strncpy (self->command, command->name, 5);
	  self->tmp_port = 0;
	  i = ftp_event (self, 1, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      switch (self->client_portpasv)
		{
		case 1:
		  if (!self->NAT)
		    {
		      if ((fakeport=ftp_get_port(self,self->client_data_address->str,0)))
			{
			  self->tmp_data_port = fakeport;
			  snprintf (buf, 29, "PORT %s,%d,%d", self->client_data_address->str, self->tmp_data_port / 256, self->tmp_data_port % 256);
			}
		      else
			{
			  // TODO: unsuccessful listen
			}
		    }
		  else
		    {
		      if (self->internal_server)
		        {	    
		          if ((fakeport=ftp_get_port(self,self->client_data_address->str,1)))
                            {
                              self->tmp_data_port = fakeport;
                              snprintf (buf, 29, "PORT %s,%d,%d", self->client_data_address->str, self->tmp_data_port / 256, self->tmp_data_port % 256);
                            }
                          else
                            {
                              // TODO: unsuccessful listen
                            }
			}
		      else
		        {
		          if ((fakeport=ftp_get_port(self,self->fw_external_data_address->str,0)))
                            {
                              self->tmp_data_port = fakeport;
                              snprintf (buf, 29, "PORT %s,%d,%d", self->fw_external_data_address->str, self->tmp_data_port / 256, self->tmp_data_port % 256);
                            }
                          else
                            {
                              // TODO: unsuccessful listen
                            }
		        }
		    }
		  g_strdelimit (buf, ".", ',');
		  ftp_command_log (self, buf, strlen (buf), command->flag);
		  break;
		case 0:
		case 2:
		  ftp_command_log (self, line, length, command->flag);
		  break;
		default:
		  break;
		}

	      self->state = FTP_SERVER_TO_CLIENT;
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	      break;
	    case Z_ABORT:
	      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	      self->quit = 1;
	      self->state = FTP_MAIN_RET;
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }

  ftp_test (self);

  return 0;
}

// Function:    ftp_command_parse_sftp
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        secure ftp extensions 
//  TODO: implementing ...
int
ftp_command_parse_sftp (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  int len;

  len = strlen (command->name);
  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((!strcmp (command->name, "CCC") && (length - len > 0)) || (strcmp (command->name, "CCC") && (length - len < 1)))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_ARG_INVALID]);
	}
      else
	{
	  strncpy (self->parameter, line + len + 1, length - len - 1);
	  self->parameter[length - len - 1] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 0, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      // AUTH ADAT PROT PBSZ MIC CONF ENC CCC
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_string
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        command with string arg
int
ftp_command_parse_string (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  int len;

  len = strlen (command->name);
  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if (length - len < 2)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_ARG_INVALID]);
	}
      else
	{
	  strncpy (self->parameter, line + len + 1, length - len - 1);
	  self->parameter[length - len - 1] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 0, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_marker
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        command with marker arg
int
ftp_command_parse_marker (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  int len;

  len = strlen (command->name);
  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if (length - len < 2)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_MARKER_INVALID]);
	}
      else
	{
	  strncpy (self->parameter, line + len + 1, length - len - 1);
	  self->parameter[length - len - 1] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 0, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_path
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        command with path arg
int
ftp_command_parse_path (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  int len;

  len = strlen (command->name);
  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((length - len - 1 > self->max_pathname_length) || (length - len < 2))
	{
	  if (!strcmp (command->name, "LIST") || !strcmp (command->name, "NLST"))
	    {
	      return ftp_command_parse_noarg (self, line, length, command);
	    }
	  else
	    {
	      ftp_command_drop (self, line, length, msgs[MSG_BAD_PATH_FORMAT]);
	    }
	}
      else
	{
	  strncpy (self->parameter, line + len + 1, length - len - 1);
	  self->parameter[length - len - 1] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 0, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      // APPE CWD RETR SMNT STOR STOU 
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_ALLO
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        command with number arg
int
ftp_command_parse_ALLO (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  int len;
  char *bug;
  char *bug2;

  len = strlen (command->name);
  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((length - len - 1 > self->max_pathname_length) || (length - len < 2))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_BAD_PATH_FORMAT]);
	}
      else
	{
	  strncpy (self->parameter, line + len + 1, length - len - 1);
	  self->parameter[length - len - 1] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);

	  strtol (self->parameter, &bug, 10);
	  if (bug - self->parameter < strlen (self->parameter))
	    {
	      if (!strncmp (bug, " R ", 3))
		{
		  strtol (bug + 3, &bug2, 10);
		  if (bug2 - self->parameter < strlen (self->parameter))
		    {
		      // bad
		      i = 9;
		    }
		  else
		    {
		      // good
		      i = Z_ACCEPT;
		    }
		}
	      else
		{
		  // bad
		  i = 9;
		}
	    }
	  else
	    {
	      // good
	      i = Z_ACCEPT;
	    }
	  if (i == Z_ACCEPT)
	    {
	      i = ftp_event (self, 0, line, length, command);
	    }
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    default:
	      ftp_command_drop (self, line, length, msgs[MSG_BAD_ALLO]);
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_USER
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        USER command
int
ftp_command_parse_USER (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  switch (self->ftp_state)
    {
    case FTP_STATE_LOGIN:
    case FTP_STATE_LOGIN_U:
    case FTP_STATE_LOGIN_P:
    case FTP_STATE_LOGIN_A:
    case FTP_STATE_LOGINAUTH:
      if (length - 6 > self->max_username_length)
	{
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: userdrop -%.*s- %d %d", self->super.session_id, length, line, length, self->max_username_length);
#endif
	  ftp_command_drop (self, line, length, msgs[MSG_BAD_USERF]);
	}
      else
	{
	  strncpy (self->username, line + 5, length - 5);
	  self->username[length - 5] = '\0';
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: username -%s-", self->super.session_id, self->username);
#endif
	  strncpy (self->command, command->name, 5);
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %08lX %08lX ", self->super.session_id, command->flag, self->command_accept);
#endif
	  i = ftp_event (self, 2, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      self->ftp_state = FTP_STATE_LOGIN_U;
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      ftp_command_drop (self, line, length, msgs[MSG_NO_USER_ACCESS]);
	    case Z_ABORT:
	      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	      self->quit = 1;
	      self->state = FTP_MAIN_RET;
	      break;
	    }
	}
      break;
    case FTP_STATE_CONVERSATION:
      ftp_command_drop (self, line, length, msgs[MSG_USER_ALREADY]);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: USER in bad place", self->super.session_id);
#endif
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_PASS
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        PASS command
int
ftp_command_parse_PASS (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  switch (self->ftp_state)
    {
    case FTP_STATE_LOGIN:
    case FTP_STATE_LOGIN_U:
    case FTP_STATE_LOGIN_P:
    case FTP_STATE_LOGIN_A:
    case FTP_STATE_LOGINAUTH:
      if (self->ftp_state != FTP_STATE_LOGIN_U)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_USER_FIRST]);
	}
      else if (length - 5 > self->max_password_length)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_BAD_PASSFORM]);
	}
      else
	{
	  strncpy (self->password, line + 5, length - 5);
	  self->password[length - 5] = '\0';
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: password -%d-", self->super.session_id, strlen (self->password));
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 2, line, length, command);
	  bzero (self->password, strlen (self->password));
	  switch (i)
	    {
	    case Z_ACCEPT:
	      self->ftp_state = FTP_STATE_LOGIN_P;
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      ftp_command_drop (self, line, length, msgs[MSG_NO_USER_ACCESS]);
	    case Z_ABORT:
	      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	      self->quit = 1;
	      self->state = FTP_MAIN_RET;
	      break;
	    }
	}
      break;
    case FTP_STATE_CONVERSATION:

      ftp_command_drop (self, line, length, msgs[MSG_USER_FIRST]);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: PASS in conversation", self->super.session_id);
#endif
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_ACCT
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        ACCT command
int
ftp_command_parse_ACCT (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  switch (self->ftp_state)
    {
    case FTP_STATE_LOGIN:
    case FTP_STATE_LOGIN_U:
    case FTP_STATE_LOGIN_P:
    case FTP_STATE_LOGIN_A:
    case FTP_STATE_LOGINAUTH:
      if ((self->ftp_state != FTP_STATE_LOGIN_U) && (self->ftp_state != FTP_STATE_LOGIN_P))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_USERPASS_FIRST]);
	}
      else if (length - 5 > self->max_password_length)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_BAD_PASSFORM]);
	}
      else
	{
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: ACCT passwd ", self->super.session_id);
#endif
	  strncpy (self->command, command->name, 5);
	  i = ftp_event (self, 2, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      self->ftp_state = FTP_STATE_LOGIN_A;
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      ftp_command_drop (self, line, length, msgs[MSG_NO_USER_ACCESS]);
	    case Z_ABORT:
	      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	      self->quit = 1;
	      self->state = FTP_MAIN_RET;
	      break;
	    }
	}
      break;
    case FTP_STATE_CONVERSATION:
      ftp_command_drop (self, line, length, msgs[MSG_USER_FIRST]);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: ACCT in conversation", self->super.session_id);
#endif
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_QUIT
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        QUIT command
int
ftp_command_parse_QUIT (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;

  switch (self->ftp_state)
    {
    case FTP_STATE_LOGIN:
    case FTP_STATE_LOGIN_U:
    case FTP_STATE_LOGIN_P:
    case FTP_STATE_LOGIN_A:
    case FTP_STATE_LOGINAUTH:
    case FTP_STATE_CONVERSATION:
      strcpy (self->parameter, "");
      strncpy (self->command, command->name, 5);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: %08lX %08lX ", self->super.session_id, command->flag, self->command_accept);
#endif
      i = ftp_event (self, 3, line, length, command);
      switch (i)
	{
	case Z_ACCEPT:
	  self->ftp_state = FTP_STATE_QUIT;
	  break;
	case Z_DENY:
	  break;
	case Z_UNSPEC:
	case Z_REJECT:
	  ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED_500]);
	case Z_ABORT:
	  ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	  self->quit = 1;
	  self->state = FTP_MAIN_RET;
	  break;
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_HELP
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        HELP command
//  TODO HELP parameter
int
ftp_command_parse_HELP (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  switch (self->ftp_state)
    {
    case FTP_STATE_LOGIN:
    case FTP_STATE_LOGIN_U:
    case FTP_STATE_LOGIN_P:
    case FTP_STATE_LOGIN_A:
    case FTP_STATE_LOGINAUTH:
    case FTP_STATE_CONVERSATION:
      strcpy (self->parameter, "");
      strncpy (self->command, command->name, 5);
      i = ftp_event (self, 0, line, length, command);
      switch (i)
	{
	case Z_ACCEPT:
	  break;
	case Z_DENY:
	  break;
	case Z_UNSPEC:
	case Z_REJECT:
	  break;
	case Z_ABORT:
	  break;
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_REIN
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        REIN command
int
ftp_command_parse_REIN (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      ftp_command_drop (self, line, length, msgs[MSG_NO_REIN]);
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: REIN ", self->super.session_id);
#endif
      // TODO: implementing? Good idea...
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_PORT
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        PORT command
int
ftp_command_parse_PORT (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;
  char newip[30];
  int newport;
  char *sp;
  unsigned int fakeport;
  //ZPolicyObj *res;

#ifdef ENABLE_DEBUG
  z_debug (7, "%s: FTP: parse PORT ", self->super.session_id);
#endif

  ftp_test (self);

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((length - 5 > 23) || (length - 5 < 1))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
	}
      else
	{
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';

	  if (strlen (self->parameter) > 24)
	    {
	      ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
	      return 1;
	    }
	  else
	    {
	      strncpy (newip, self->parameter, 24);

	      if (strtol (newip, NULL, 10) > 255)
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
		  return 1;
		}
      if ( errno == ERANGE )
        {
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
	  return 1;
	}
	      for (i = 0; i < strlen (newip); i++)
		{
		  if (!isdigit (newip[i]))
		    {
		      if (newip[i] == ',')
			{
			  if (strtol (newip + i + 1, NULL, 10) > 255)
			    {
			      ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
			      return 1;
			    }
      if ( errno == ERANGE )
        {
			      ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
	  return 1;
	}
			  newip[i] = '.';
			}
		      else
			{
			  break;
			}
		    }
		}
	      if (i != strlen (newip))
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_PORT]);
		  return 1;
		}

	      sp = strrchr (newip, '.');
	      *sp = ',';
	      sp = strrchr (newip, '.');
	      *sp = '\0';
	      sp++;
	      newport = strtol (sp, NULL, 10);
	      newport *= 256;
	      sp = strrchr (sp, ',');
	      sp++;
	      newport += strtol (sp, NULL, 10);

	      g_string_assign (self->tmp_data_address, newip);
	      self->tmp_data_port = newport;

	      if ( (!self->internal_server) && (self->NAT) )
		{
		  if ((fakeport=ftp_get_port(self,self->fw_external_data_address->str,0)))
		    {
		      self->tmp2_data_port = fakeport;
		      snprintf (newip, 29, "PORT %s,%d,%d", self->fw_external_data_address->str, self->tmp2_data_port / 256, self->tmp2_data_port % 256);
		      g_strdelimit (newip, ".", ',');
		    }
		  else
		    {
		      // TODO: unsuccessful listen
		    }
		}

	    }

#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  self->tmp_port = 1;

	  if (!ftp_command_flag (self, 0, command->flag, self->command_accept))
	    {
	      z_policy_lock (self->super.thread);
	      i = z_policy_event (self->super.handler, command->event, z_policy_var_build ("clsssi", ftp_command_getCC (command->flag), command->flag, self->command, self->parameter, self->tmp_data_address->str, self->tmp_data_port));
	      z_policy_unlock (self->super.thread);
	    }
	  else
	    {
	      i = Z_ACCEPT;
	    }
	  //i = ftp_event (self, line, length, command);
	  switch (i)
	    {
	    case Z_ACCEPT:
	      switch (self->client_portpasv)
		{
		case 2:
		  ftp_command_log (self, "PASV", strlen ("PASV"), command->flag);
		  break;
		case 0:
		case 1:
	          if ( (!self->internal_server) && (self->NAT) )
		    {
		      ftp_command_log (self, newip, strlen (newip), command->flag);
		    }
		  else
		    {
		      ftp_command_log (self, line, length, command->flag);
		    }
		  break;
		default:
		  break;
		}
	      self->state = FTP_SERVER_TO_CLIENT;

	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	      break;
	    case Z_ABORT:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }

  ftp_test (self);

  return 0;
}

// Function:    ftp_command_parse_TYPE
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        TYPE command
int
ftp_command_parse_TYPE (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if ((length - 5 > 3) || (length - 5 < 1) || (length - 5 == 2))
	{
	  ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
	}
      else
	{
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
	  z_debug (7, "%s:   %08lX %08lX ", self->super.session_id, command->flag, (unsigned long) self->command_accept);
#endif
	  strncpy (self->command, command->name, 5);
	  i = Z_REJECT;

#ifdef ENABLE_DEBUG
	  z_debug (7, "%s:   TYPE - parser ", self->super.session_id);
#endif
	  switch (toupper (self->parameter[0]))
	    {
	    case 'I':
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s:   TYPE - parser I", self->super.session_id);
#endif
	      if (strlen (self->parameter) > 1)
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
		  i = 9;
		}
	      else
		{
		  i = Z_ACCEPT;
		}
	      break;
	    case 'L':
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s:   TYPE - parser L %d -%c-", self->super.session_id, strlen (self->parameter), self->parameter[1]);
#endif
	      if ((strlen (self->parameter) != 3) || (self->parameter[1] != ' '))
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
		  i = 9;
		}
	      else if (self->parameter[2] != '8')
		{
		  ftp_command_drop (self, line, length, msgs[MSG_TYPE_BYTE_SIZE]);
		  i = 9;
		}
	      else
		{
		  i = Z_ACCEPT;
		}
	      break;
	    case 'A':
#ifdef ENABLE_DEBUG
	      z_debug (7, "%s:   TYPE - parser A", self->super.session_id);
#endif
	      if ((strlen (self->parameter) != 1) && (strlen (self->parameter) != 3))
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
		  i = 9;

		}
	      if ((strlen (self->parameter) == 3) && (self->parameter[1] != ' '))
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
		  i = 9;
		}
	      else if ((strlen (self->parameter) == 3) && (toupper (self->parameter[2]) != 'N'))
		{
		  ftp_command_drop (self, line, length, msgs[MSG_TYPE_FORM]);
		  i = 9;
		}
	      else
		{
		  i = Z_ACCEPT;
		}
	      break;
	    case 'E':
	      if (strlen (self->parameter) != 1)
		{
		  ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
		  i = 9;

		}
	      else
		{
		  ftp_command_drop (self, line, length, msgs[MSG_TYPE_E]);
		  i = 9;
		}
	      break;
	    default:
	      ftp_command_drop (self, line, length, msgs[MSG_CANT_TYPE]);
	      i = 9;
	      break;
	    }
	  if ((i == Z_ACCEPT))
	    {
	      i = ftp_event (self, 4, line, length, command);
	    }
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	      break;
	      break;
	    case Z_ABORT:
	      break;
	    case 9:
	      ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	      break;
	    default:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_STRU
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        STRU command
int
ftp_command_parse_STRU (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
      if (length - 5 != 1)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_INVALID_STRU]);
	}
      else
	{
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  i = Z_ACCEPT;
	  switch (self->parameter[0])
	    {
	    case 'F':
	    case 'f':
	      break;
	    case 'R':
	    case 'P':
	    case 'r':
	    case 'p':
	      ftp_command_drop (self, line, length, msgs[MSG_UNIMPLEMENTED_STRU]);
	      i = 9;
	      break;
	    default:
	      ftp_command_drop (self, line, length, msgs[MSG_INVALID_STRU]);
	      i = 9;

	    }
	  if (i == Z_ACCEPT)
	    {
	      i = ftp_event (self, 0, line, length, command);
	    }
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	    case Z_ABORT:
	      break;
	    case 9:
	      ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	      break;
	    default:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_MODE
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              FTPCommandDescriptor    found command
// Out:         int
// Desc:        MODE command
int
ftp_command_parse_MODE (FtpProxy * self, char *line, int length, FTPCommandDescriptor * command)
{
  int i;

  switch (self->ftp_state)
    {

    case FTP_STATE_CONVERSATION:
      if (length - 5 != 1)
	{
	  ftp_command_drop (self, line, length, msgs[MSG_INVALID_MODE]);
	}
      else
	{
	  strncpy (self->parameter, line + 5, length - 5);
	  self->parameter[length - 5] = '\0';
#ifdef ENABLE_DEBUG
	  z_debug (7, "%s: FTP: %s parameter -%s-", self->super.session_id, command->name, self->parameter);
#endif
	  strncpy (self->command, command->name, 5);
	  i = Z_ACCEPT;
	  switch (self->parameter[0])
	    {
	    case 'S':
	    case 's':
	      break;
	    case 'B':
	    case 'C':
	    case 'b':
	    case 'c':
	      ftp_command_drop (self, line, length, msgs[MSG_UNIMPLEMENTED_MODE]);
	      i = 9;
	      break;
	    default:
	      ftp_command_drop (self, line, length, msgs[MSG_INVALID_MODE]);
	      i = 9;

	    }
	  if (i == Z_ACCEPT)
	    {
	      i = ftp_event (self, 0, line, length, command);
	    }
	  switch (i)
	    {
	    case Z_ACCEPT:
	      break;
	    case Z_DENY:
	      break;
	    case Z_UNSPEC:
	    case Z_REJECT:
	      break;
	    case Z_ABORT:
	      break;
	    case 9:
	      ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	      break;
	    default:
	      break;
	    }
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}

// Function:    ftp_command_parse_UNKN
// In:          FtpProxy                Proxy pointer
//              char*                   sended line
//              int                     length of line
//              long                    free  flag
// Out:         int
// Desc:        unknown command
int
ftp_command_parse_UNKN (FtpProxy * self, char *line, int length, long flag)
{
  int i;

  switch (self->ftp_state)
    {
    case FTP_STATE_CONVERSATION:
#ifdef ENABLE_DEBUG
      z_debug (7, "%s: FTP: unknown command", self->super.session_id);
#endif
      strncpy (self->command, line, 4);
      self->command[4] = '\0';
      if (!self->command_unk)
	{
	  z_policy_lock (self->super.thread);
	  i = z_policy_event (self->super.handler, "unknownCommand", z_policy_var_build ("css#", 'C', self->command, line, length));
	  z_policy_unlock (self->super.thread);
	}
      else
	{
	  i = Z_REJECT;
	}
      switch (i)
	{
	case Z_ACCEPT:
	  ftp_command_log (self, line, length, flag);
	  self->state = FTP_SERVER_TO_CLIENT;
	  break;
	case Z_DENY:
	  break;
	case Z_UNSPEC:
	case Z_REJECT:
	  ftp_command_drop (self, line, length, msgs[MSG_COMMAND_DISABLED]);
	case Z_ABORT:
	  ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
	  self->quit = 1;
	  self->state = FTP_MAIN_RET;
	  break;
	default:
	  break;
	}
      break;
    default:
      ftp_command_drop (self, line, length, msgs[MSG_Z_ABORT]);
      self->quit = 1;
      self->state = FTP_MAIN_RET;
      break;
    }
  return 0;
}
