/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: http.h,v 1.10 2000/08/08 14:47:17 bazsi Exp $
 *
 * By: Kovacs Viktor Peter (vps__@hotmail.com | vps__@freemail.hu | ...)
 *
 ***************************************************************************/
#include <zorp/zorp.h>
#include <zorp/stream.h>
#include <zorp/proxy.h>
#include <zorp/policy.h>
#include <zorp/thread.h>
#include <zorp/zpython.h>
#include <zorp/log.h>
#include <zorp/registry.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>  	      
#include "glib.h"
#include "httpstr.h"
#include "httptool.h"
#include "http_def.h"
#include "httpprot.h"


#define HTTP_ERROR (-1)

// #define TRUE  (1)
// #define FALSE (0)

#define HTTP_BLOCKSIZE   (8192)  //(512)
#define HTTP_LOGLINESIZE (1024*10)

enum HTTP_SIDE
{
 HTTP_SIDE_CLIENT=1,
 HTTP_SIDE_SERVER,
 HTTP_SIDE_BOTH
};

enum HTTP_ERR           //levels: (4 error, 4 event) ; 0=highest
{
 HTTP_ERR_INTERNAL=1,   //0
 HTTP_ERR_FATAL,        //0
 HTTP_ERR_UNKNOWN,      //0
 HTTP_ERR_CLIENT_FAIL,  //3
 HTTP_ERR_SERVER_FAIL,  //2
 HTTP_ERR_CONNECT,      //2
 HTTP_ERR_CLIENT_SYNTAX,//2
 HTTP_ERR_SERVER_SYNTAX,//2

 HTTP_ERR_POLICY,       //0

 HTTP_EVENT_INIT,       //0
 HTTP_EVENT_SHUTDOWN,   //0

 HTTP_EVENT_ACCEPT,     //1
 HTTP_EVENT_CONNECT,    //1
 HTTP_EVENT_CLOSE_CLIENT_IMMINENT, //1
 HTTP_EVENT_CLOSE_CLIENT_GRACEFUL, //1
 HTTP_EVENT_CLOSE_SERVER_IMMINENT, //1
 HTTP_EVENT_CLOSE_SERVER_GRACEFUL, //1
 HTTP_EVENT_FILTER,      //2
 HTTP_EVENT_FILTER_IN,   //3
 HTTP_EVENT_FILTER_OUT,  //3
 HTTP_EVENT_POLICY_CALL, //3

 HTTP_EVENT_REQUEST,     //2
 HTTP_EVENT_RESPONSE,    //2
 HTTP_EVENT_COMPLETE     //1
};

enum HTTP_LOG_MODE
{
 HTTP_LOG_ERR0=128,
 HTTP_LOG_ERR1=64,
 HTTP_LOG_ERR2=32,
 HTTP_LOG_ERR3=16,
 HTTP_LOG_EVN0=8,
 HTTP_LOG_EVN1=4,
 HTTP_LOG_EVN2=2,
 HTTP_LOG_EVN3=1
};

enum HTTP_EFID
{
 HTTP_EFID_INTERNAL=1,
 HTTP_EFID_SYNTAX,
 HTTP_EFID_CONNECT,
 HTTP_EFID_SERVER,
 HTTP_EFID_POLICY
};

enum HTTP_MODE
{
 HTTP_MODE_CLOSE=1,
 HTTP_MODE_KEEPALIVE
};

enum HTTP_FILTER
{
 HTTP_FILTER_PASS=1,
 HTTP_FILTER_DROP,
 HTTP_FILTER_CHANGE_BOTH,
 HTTP_FILTER_CHANGE_NAME,
 HTTP_FILTER_CHANGE_VALUE,
 
 HTTP_FILTER_CHANGE_REGEXP, //TODO
 HTTP_FILTER_INSERT,        //TODO

 HTTP_FILTER_POLICY
};

enum HTTP_FILTER_WHAT
{
 HTTP_FILTER_WHAT_REQ_CMD=1,
 HTTP_FILTER_WHAT_REQ_HEAD,
 HTTP_FILTER_WHAT_RESP_CMD,
 HTTP_FILTER_WHAT_RESP_HEAD
};

typedef struct _HttpProxy{
 ZProxy super;
 int cstat;       //connection status: client
 int sstat;       //connection status: server
 int mode;        //0:close 1:keep-alive
 int transp;      //0:non-transparent 1:transparent
 int ver;         //version (updated by request and response) : 0x major minor (0x0009 0x0100 0x0101)
 int logmode;     //bitmapped: 0b .... .... : err0 err1 err2 err3  evn0 evn1 evn2 evn3
 int maxreqline;  //max. number of lines in the request (command+headers)
 int maxrespline; //max. number of lines in the response (result+headers)
 int maxlinelen;  //max. line length
 int reqflg;      //request command flags
 int respflg;     //response value flags
 GHashTable* req_cmd;                     //command and header lookup hashes
 GHashTable* req_head;
 GHashTable* resp_cmd;
 GHashTable* resp_head;
 int req_cmd_def;                         //default action to take when not found above
 int req_head_def;                        //allowed: PASS | DROP
 int resp_cmd_def;
 int resp_head_def;
 GHashTable* filter_transfer;             //filter eslection tables (unused)
 GHashTable* filter_encoding;
 GHashTable* filter_content;
 void* filter_transfer_data;              //save area for filters
 void* filter_encoding_data;
 void* filter_content_data;
 int   (*filter_transfer_error)(void* s); //default error handler for filters
 int   (*filter_encoding_error)(void* s);
 int   (*filter_content_error)(void* s); 
 HttpStr* addr;        //address of the server
 HttpStr* old_addr;    //reconnect if !=addr   => this should contain the name of the connected address 
 int      transp_addr; //connect with transparent chainer (addr ignored)
 char bufc[HTTP_BLOCKSIZE];  //client read buffer
 char bufs[HTTP_BLOCKSIZE];  //server read buffer
 int bufcpos;              //client buffer position
 int bufclen;              //-||-   buffer length
 int bufspos;              //server -||-
 int bufslen;
}HttpProxy;

//public functions used by other parts of the proxy
int http_recv(HttpProxy* s, int side);
int http_send(HttpProxy* s, int side, char* buf, int len);
int http_read(HttpProxy* s, int side, HttpStr* buf);
int http_write(HttpProxy* s, int side, HttpStr* buf);
int http_readline(HttpProxy* s, int side, HttpStr* buf, int maxlen);
int http_writeline(HttpProxy* s, int side, HttpStr* buf);

int http_less(int a, int b);
